#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: utils.py
# Author: Rayson.Shi
# Mail: raysonshi@qq.com
# Created Time:  2022-8-5 23:17:34
#############################################
import os
import numpy as np
import cv2
import json
import copy
import datetime


LICENSE = {
"id": None, "name": None, "url": None,
}


COCO_FORMAT  = {
"info": [], "images": [], "annotations": [], "categories":[], "licenses": LICENSE
}


INFO = {
"year": None, "version": None, "description": None, "contributor": None, "url": None, "date_created": None,
}


IMAGE = {
"id": None, "width": None, "height": None, "file_name": None, "license": None, "flickr_url": None, "coco_url": None, "date_captured": None,
}


ANNOTATION = {
"id": None, "image_id": None, "category_id": None, "segmentation": None, "area": None, "bbox": None, "iscrowd": 0,
}


CATEGORIE = {
"id": None, "name": None, "supercategory": None,
}


def xywh2xyxy(xywh):
    x2 = xywh[2] + xywh[0]
    y2 = xywh[3] + xywh[1]
    return [xywh[0], xywh[1], x2, y2]


def xyxy2xywh(xyxy):
    w = xyxy[2] - xyxy[0]
    h = xyxy[3] - xyxy[1]
    return [xyxy[0], xyxy[1], w, h]


def getAndCheck(root, name, length=None):
    vals = root.findall(name)
    if len(vals) == 0:
        raise NotImplementedError('Can not find %s in %s.' % (name, root.tag))
    if length and len(vals) != length:
        raise NotImplementedError('The size of %s is supposed to be %d, but is %d.' % (name, length, len(vals)))
    if length == 1:
        vals = vals[0]
    return vals


def checkImgEists(name, formats=[".jpg", ".png"]):
    for f in formats:
        path = os.path.join(name, f)
        if os.path.exists(path):
            return path
    return None


def setImageSample(img_id, img_name, h, w):
    img_sample = copy.deepcopy(IMAGE)
    img_sample["id"] = img_id
    img_sample["file_name"] = img_name
    img_sample["height"] = h
    img_sample["width"] = w
    return img_sample


def setAnnotationSample(ano_id, img_id, cat_id, bbox, polygon=None):
    ano_sample = copy.deepcopy(ANNOTATION)
    ano_sample["id"] = ano_id
    ano_sample["image_id"] = img_id
    ano_sample["category_id"] = cat_id
    ano_sample["bbox"] = list(map(int, bbox))
    ano_sample["area"] = bbox[2] * bbox[3]
    ano_sample["segmentation"] = polygon
    return ano_sample


def setCategorySample(cat, cat_names):
    if cat in cat_names:
        return cat_names.index(cat),None
    else:
        cat_id = len(cat_names)
        cat_sample = copy.deepcopy(CATEGORIE)
        cat_sample["id"] = cat_id
        cat_sample["name"] = cat
        cat_sample["supercategory"] = None
        cat_names.append(cat)
        return cat_id, cat_sample


def points2xyxy(points):
    points = np.array(points).reshape(-1, 2)
    x1 = np.min(points[:, 0])
    x2 = np.max(points[:, 0])
    y1 = np.min(points[:, 1])
    y2 = np.max(points[:, 1])
    return [x1, y1, x2, y2]


def yoloBox2xyxy(yolo_box, img_h, img_w):
    center_x = int(img_w * yolo_box[0])
    center_y = int(img_h * yolo_box[1])
    box_w = int(img_w * yolo_box[2])
    box_h = int(img_h * yolo_box[3])
    x1 = center_x - int(box_w/2)
    x2 = center_x + int(box_w/2)
    y1 = center_y - int(box_h/2)
    y2 = center_y + int(box_h/2)
    return [x1, y1, x2, y2]


def checkXyxy(xyxy, img_w, img_h):
    x1, y1, x2, y2 = xyxy
    if any([x1<0, y1<0, x2<0, y2<0]):
        print("illegal input xyxy lower than 0: {},{},{},{}".format(x1,y1,x2,y2))
        return None
    if any([x1>img_w, x2>img_w]) or any([y1>img_h, y2>img_h]):
        print("illegal input xyxy exceeds picture size: {},{},{},{}".format(x1, y1, x2, y2))
        return None
    if any([x2-x1<2, y2-y1<2]):
        print("illegal input xyxy: {},{},{},{}".format(x1, y1, x2, y2))
        return None
    return xyxy


def repairBox(xywh, img_h, img_w):
    xyxy = xywh2xyxy(xywh)
    x1 = np.clip(xyxy[0], 1, img_w-1)
    x2 = np.clip(xyxy[2], 1, img_w-1)
    y1 = np.clip(xyxy[1], 1, img_h-1)
    y2 = np.clip(xyxy[3], 1, img_h-1)
    xywh = xyxy2xywh([x1, y1, x2, y2])
    return xywh


def drawBox(img, xywh, color):
    xyxy = xywh2xyxy(xywh)
    cv2.rectangle(img, tuple(xyxy[:2]), tuple(xyxy[2:]), color, 2)


def drawTxt(img, txt, point, color):
    cv2.putText(img, txt, tuple(point), cv2.FONT_HERSHEY_SIMPLEX, 0.75, tuple(color), 2)


def drawPolygons(img, polygons, color, alpha=0.7):
    mask = img.copy()
    cv2.fillPoly(mask, [polygons], color)
    add_img = cv2.addWeighted(img, alpha, mask, 1-alpha)
    return add_img


def drawKeypoints(img, keypoints, skeleton, point_color=[0, 0, 255], line_color=[0, 255, 0]):
    for sk in skeleton:
        pt1 = keypoints[sk[0]]
        pt2 = keypoints[sk[1]]
        cv2.line(img, pt1, pt2, line_color, 2)
    for p in keypoints:
        cv2.circle(img, p, 4, point_color)


def checkXywh(wrong_ano_number, img_id, ano_id, xywh, img_w, img_h, log):
    xyxy = xywh2xyxy(xywh)
    x1, y1, x2, y2 = xyxy
    if any([x1<0, y1<0, x2<0, y2<0]):
        log.info("img id: {} ano id: {}  xywh: {} img w: {} img h: {} result: Invalid values!".format(img_id, ano_id, xywh, img_w, img_h))
        wrong_ano_number += 1
    if any([x1>img_w, x2>img_w]) or any([y1>img_h, y2>img_h]):
        log.info("img id: {} ano id: {}  xywh: {} img w: {} img h: {} result: Over the border!".format(img_id, ano_id, xywh, img_w, img_h))
        wrong_ano_number += 1
    if any([x2-x1<2, y2-y1<2]):
        log.info("img id: {} ano id: {}  xywh: {} img w: {} img h: {} result: Invalid values!".format(img_id, ano_id, xywh, img_w, img_h))
        wrong_ano_number += 1
    return wrong_ano_number


class MyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, bytes):
            return str(obj, encoding='utf-8')
        elif isinstance(obj, datetime.datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        else:
            return super(MyEncoder, self).default(obj)



#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: analyzer.py
# Author: Rayson.Shi
# Mail: raysonshi@qq.com
# Created Time:  2022-8-5 23:17:34
#############################################

import collections
import numpy as np
from pycocotools.coco import COCO
from cocoutils.drwaUtils import DrawUtils


class Analyzer:
    def __init__(self, coco_path, save_path=None, map_size=8):
        """
        Analyze coco datasets through charts
        :param coco_path (str) : the path of Coco format JSON file
        :param save_path (dir) : the dir to save the chart, if None, it will display directly on the desktop
        :param map_size (int)  : Determine the number of heatmap grids
        """
        super(Analyzer, self).__init__()
        self.coco = COCO(coco_path)
        self.map_size = map_size
        self.s_threshold = 32 * 32
        self.m_threshold = 96 * 96
        self.__drawUtil = DrawUtils(save_path)
        self.__create_data()
        self.__draw()

    def __create_data(self):
        img_ids = self.coco.getImgIds()
        ano_ids = self.coco.getAnnIds()
        cat_ids = self.coco.getCatIds()
        imgs_num = len(img_ids)
        anos_num = len(ano_ids)
        cat_num = len(cat_ids)
        self.imgs_data = np.zeros((imgs_num, 4))
        self.anos_data = np.zeros((anos_num, 8))
        self.cats_data = np.zeros((cat_num, 2))
        for ig_id in img_ids:
            img_sample = self.coco.loadImgs(ig_id)[0]
            self.imgs_data[ig_id, 0] = img_sample["width"]
            self.imgs_data[ig_id, 1] = img_sample["height"]
            img_area = img_sample["width"] * img_sample["height"]
            relate_anos_ids = self.coco.getAnnIds(ig_id)
            self.imgs_data[ig_id, 3] = len(relate_anos_ids)
            for ano_id in relate_anos_ids:
                ano_sample = self.coco.loadAnns(ano_id)[0]
                self.anos_data[ano_id, 0] = ano_sample["area"]
                self.anos_data[ano_id, 1] = self.__calObjSize(ano_sample["area"], self.s_threshold, self.m_threshold)
                self.anos_data[ano_id, 2] = ano_sample['area'] / img_area
                self.anos_data[ano_id, 3] = ano_sample["bbox"][3] / ano_sample["bbox"][2]
                self.anos_data[ano_id, 4] = ano_sample["category_id"]
                self.anos_data[ano_id, 5], self.anos_data[ano_id, 6] = self.__calMapIndex(ano_sample["bbox"], \
                                                                                   self.map_size,      \
                                                                                   img_sample["height"], \
                                                                                   img_sample["width"])
                self.anos_data[ano_id, 7] = ano_sample["image_id"]
        for c_id in cat_ids:
            self.cats_data[:, 0] = self.coco.getImgIds(catIds=[c_id])[0]
            self.cats_data[:, 1] = sum(self.anos_data[:, 4] == c_id)

    def __draw(self):
        overall_labels, overall_values = self.__calOverallInformation()
        self.__drawUtil.drawBar("General information", "name", "number", overall_labels, overall_values)

        img_size_labels, img_size_values = self.__calImgSize()
        self.__drawUtil.drawBar("Image size number", "image size", "image number", img_size_labels, img_size_values, 10)

        obj_number_labels, img_number_values = self.__calImgObjNumber()
        self.__drawUtil.drawBar("Number of targets per picture", "object number", "image number", obj_number_labels, img_number_values)

        cats_labels, cats_values = self.__calCategoryNumber()
        self.__drawUtil.drawPie("Quantity proportion of each category", cats_labels, cats_values)
        self.__drawUtil.drawBar("Number of each category", "category name", "category number", cats_labels, cats_values)

        cat_labels, img_values = self.__calCatImgNumber()
        self.__drawUtil.drawBar("Number of pictures in each category", "category name", "image number",
                              cat_labels, img_values)

        size_label, size_values = self.__calAbsArea()
        self.__drawUtil.drawBar("Absolute target area", "object size", "object number", size_label, size_values)

        relative_size = self.__calRealateSize()
        self.__drawUtil.drawHist("Object relative area distribution", "relative size", "number", relative_size)

        object_areas = self.anos_data[:, 0]
        self.__drawUtil.drawHist("Object area distribution", "area", "number", object_areas, 100)

        map_values = self.__calMapDist()
        labels = [str(i) for i in range(self.map_size)]
        self.__drawUtil.drawHeatmap("Object position distribution", labels, labels, map_values)

        aspect_ratio = self.__calAspectRatio()
        self.__drawUtil.drawHist("Aspect ratio distribution", "Aspect ratio", "number", aspect_ratio)

    def __calOverallInformation(self):
        labels = ["image number", "annotation number", "category number"]
        values = [self.imgs_data.shape[0], self.anos_data.shape[0], self.cats_data.shape[0]]
        return labels, values

    def __calImgSize(self):
        labels = []
        values = []
        unique_ws = np.unique(self.imgs_data[:, 0])
        for u_w in unique_ws:
            realate_indexs = np.where(self.imgs_data[:, 0]==u_w)
            realte_hs = self.imgs_data[:, 1][realate_indexs]
            unique_hs = np.unique(realte_hs)
            for u_h in unique_hs:
                hs = np.where(realte_hs==u_h)[0]
                la = "(" + str(u_w) + "," + str(u_h) +")"
                labels.append(la)
                val = hs.size
                values.append(val)
        return labels, values

    def __calImgObjNumber(self):
        data = collections.Counter(self.imgs_data[:, 3])
        labes = [int(x) for x in list(data.keys())]
        values = list(data.values())
        return labes, values

    def __calCategoryNumber(self):
        labels = []
        values = []
        cat_ids = self.coco.getCatIds()
        for c_i in cat_ids:
            la = self.coco.loadCats(c_i)[0]["name"]
            val = np.where(self.anos_data[:, 4]==c_i)[0].size
            labels.append(la)
            values.append(val)
        return labels, values

    def __calAbsArea(self):
        labels = ["small", "medium", "big"]
        values = []
        for i in range(3):
            val = np.where(self.anos_data[:, 1]==i)[0].size
            values.append(val)
        return labels, values

    def __calRealateSize(self):
        return self.anos_data[:, 2]

    def __calMapDist(self):
        values = np.zeros((self.map_size, self.map_size))
        for i in range(self.map_size):
            for j in range(self.map_size):
                x_indexs = np.where(self.anos_data[:, 5] == i)
                r_y = self.anos_data[:, 6][x_indexs]
                val = np.where(r_y==j)[0].size
                values[i, j] = val
        return values

    def __calAspectRatio(self):
        return self.anos_data[:, 3]

    def __calCatImgNumber(self):
        values = []
        cat_ids = self.coco.getCatIds()
        label = [x["name"] for x in self.coco.loadCats(cat_ids)]
        for c_i in cat_ids:
            index = np.where(self.anos_data[:, 4] == c_i)
            related_img_id = self.anos_data[:, 7][index]
            val = np.unique(related_img_id).size
            values.append(val)
        return label, values

    def __calObjSize(self, area, s_thredshold, m_threshold):
        return 0 if area < s_thredshold else (1 if area < m_threshold else 2)

    def __calMapIndex(self, bbox, map_size, img_h, img_w):
        center_x = bbox[0] + int(bbox[2] / 2)
        center_y = bbox[1] + int(bbox[3] / 2)
        x_index = int((center_x / img_w) * map_size)
        y_index = int((center_y / img_h) * map_size)
        return x_index, y_index

if __name__=="__main__":
    coco = r"coco.json"
    analysis_result = r"analysis_result"
    Analyzer(coco, analysis_result)





#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: drwaUtils.py
# Author: Rayson.Shi
# Mail: raysonshi@qq.com
# Created Time:  2022-8-5 23:17:34
#############################################
import os
import matplotlib.pyplot as plt
import numpy as np


class DrawUtils():
    def __init__(self, save_path):
        super(DrawUtils, self).__init__()
        self.save_path = save_path
        if self.save_path and not os.path.exists(save_path):
            os.mkdir(self.save_path)
        plt.rcParams['font.sans-serif'] = 'SimHei'  # 设置中文显示
        plt.rcParams["axes.unicode_minus"] = False

    def drawPie(self, title, labels, values):
        indexs = list(reversed(np.argsort(values)))
        labels = [labels[i] for i in indexs]
        values = [values[i] for i in indexs]
        explode = [0.01 for _ in range(len(labels))]
        plt.pie(values, explode=explode, labels=labels, autopct='%1.1f%%')
        plt.title(title)
        if self.save_path:
            save_path = os.path.join(self.save_path, title + ".jpg")
            plt.savefig(save_path)
            plt.close()
        else:
            plt.show()
            plt.close()

    def drawHeatmap(self, title, x_lables, y_labels, map_vales):
        plt.xticks(np.arange(len(x_lables)), labels=x_lables, rotation_mode="anchor", ha="right")
        plt.yticks(np.arange(len(y_labels)), labels=y_labels)
        for i in range(len(x_lables)):
            for j in range(len(y_labels)):
                plt.text(j, i, int(map_vales[i, j]), ha="center", va="center", color="w")
        plt.title(title)
        plt.imshow(map_vales)
        plt.colorbar()
        plt.tight_layout()
        if self.save_path:
            save_path = os.path.join(self.save_path, title + ".jpg")
            plt.savefig(save_path)
            plt.close()
        else:
            plt.show()
            plt.close()

    def drawBar(self, title, x_name, y_name, labels, values, max_number=None):
        indexs = list(reversed(np.argsort(values)))
        labels = [labels[i] for i in indexs]
        values = [values[i] for i in indexs]
        if max_number and (len(labels) > max_number):
            labels = labels[:max_number]
            values = values[:max_number]
            print("{} chart only show the top {}".format(title, max_number))
        plt.bar(range(len(labels)), values)
        if type(labels[0])==str and len(labels) > 8:
            plt.xticks(range(len(labels)), labels, rotation=-45)
        else:
            plt.xticks(range(len(labels)), labels)
        for i in range(len(labels)):
            plt.text(i, values[i], '%.i' % values[i], fontsize=10, ha='center', va='bottom')
        plt.xlabel(x_name)
        plt.ylabel(y_name)
        plt.title(title)
        if self.save_path:
            save_path = os.path.join(self.save_path, title + ".jpg")
            plt.savefig(save_path)
            plt.close()
        else:
            plt.show()
            plt.close()

    def drawHist(self, title, x_names, y_name, values, bins=100):
        plt.hist(values, bins=bins, facecolor="blue", edgecolor="black", alpha=0.7)
        plt.xlabel(x_names)
        plt.ylabel(y_name)
        plt.title(title)
        if self.save_path:
            save_path = os.path.join(self.save_path, title + ".jpg")
            plt.savefig(save_path)
            plt.close()
        else:
            plt.show()
            plt.close()




#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: checker.py
# Author: Rayson.Shi
# Mail: raysonshi@qq.com
# Created Time:  2022-8-5 23:17:34
#############################################
import random
from pycocotools.coco import COCO
import logging
from cocoutils.utils import *


class Checker():
    def __init__(self, coco_path):
        """
        Check dataset outliers and visualization
        :param coco_path (str) : the path of Coco format JSON file
        """
        super(Checker, self).__init__()
        self.coco = COCO(coco_path)
        random.seed(123)
        self.__cat_colors = [[random.randint(0, 255), random.randint(0, 255), random.randint(0, 255)] \
                           for _ in self.coco.cats]

    def drawImg(self, img_dir, img_id, save_dir=None):
        """
        Draw annotations through image index
        :param img_dir (str) : picture save folder
        :param img_id (int) : image id
        :param save_dir (int) : save the picture folder after drawing, if None, it will display directly on the desktop
        """
        img_sample = self.coco.loadImgs(img_id)[0]
        img_name = img_sample["file_name"]
        img_path = os.path.join(img_dir, img_name)
        img = cv2.imread(img_path)
        ano_ids = self.coco.getAnnIds(img_id)
        for a_i in ano_ids:
            ano_sample = self.coco.loadAnns(a_i)[0]
            cat_id = ano_sample["category_id"]
            cat_sample = self.coco.loadCats(cat_id)[0]
            if ano_sample["bbox"] is not None:
                if (len(ano_sample["bbox"]) > 0):
                    drawBox(img, ano_sample["bbox"], self.__cat_colors[cat_id])
                    bt = ano_sample["bbox"][:2]
                    bt[1] = bt[1] - 5
                    drawTxt(img, cat_sample["name"], bt, (0,245,0))
        if save_dir:
            save_path = os.path.join(save_dir, img_name)
            cv2.imwrite(save_path, img)
            print("Saving img to {}".format(save_path))
        else:
            cv2.imshow(img_name, img)
            cv2.waitKey(0)
            cv2.destroyAllWindows()

    def check(self, img_dir=None, log="check.txt"):
        """
        Check annotations and pictures
        :param img_dir (str) : picture save folder, if None, do not check whether the picture exists
        :param log (path) : address to save inspection results
        """
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        fh = logging.FileHandler(log, mode='w')
        fh.setLevel(logging.INFO)
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        formatter = logging.Formatter("%(message)s")
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        logger.addHandler(fh)
        logger.addHandler(ch)
        img_ids = self.coco.getImgIds()
        unique_img_ids = set(img_ids)
        ano_ids = self.coco.getAnnIds()
        unique_ano_ids = set(ano_ids)
        img_names = []
        cat_ids = self.coco.getCatIds()
        cats = self.coco.loadCats(cat_ids)
        wrong_ano_number = 0
        assert len(img_ids) > 0, "the images is empty!"
        assert len(ano_ids) > 0, "the annotations is empty!"
        assert len(cats) > 0, "the categories is empty!"
        for ig_id in img_ids:
            img_sample = self.coco.loadImgs(ig_id)[0]
            img_w = img_sample["width"]
            img_h = img_sample["height"]
            img_names.append(img_sample["file_name"])
            if img_dir:
                img_path = os.path.join(self.img_dir, img_sample["file_name"])
                if not os.path.exists(img_path):
                    logger.info("img id: {} path {} result: Path does not exist! ".format(ig_id, img_path))
            t_ano_ids = self.coco.getAnnIds(ig_id)
            if len(t_ano_ids) == 0:
                logger.info("img id: {} result: No annotation information! ".format(ig_id, img_path))
                wrong_ano_number += 1
            for ano_id in t_ano_ids:
                ano_sample = self.coco.loadAnns(ano_id)[0]
                cat_id = ano_sample["category_id"]
                if cat_id not in cat_ids:
                    logger.info("img id: {} ano id: {} category id: {} result: No annotation information!".format(ig_id, ano_id, cat_id))
                    wrong_ano_number += 1
                xywh = ano_sample["bbox"]
                wrong_ano_number = checkXywh(wrong_ano_number, ig_id, ano_id, xywh, img_w, img_h, logger)
        if len(unique_ano_ids) != len(ano_ids):
            logger.info(
                "Mismatch! the total ano id number is: {}, remove duplicate ano number is {}" \
                  .format(len(ano_ids), len(unique_ano_ids)))
        if len(unique_img_ids) != len(img_ids):
            logger.info(
                "Mismatch! the total img id number is: {}, remove duplicate img number is {}" \
                  .format(len(img_ids), len(unique_img_ids)))
        unique_img_names = set(img_names)
        if len(unique_img_names) != len(img_names):
            logger.info(
                "Mismatch! the total img name number is: {},remove duplicate img name number is {}" \
                  .format(len(img_names), len(unique_img_names)))
        print("Total annotation number is {}, the wrong annotation number is {}".format(len(ano_ids), wrong_ano_number))
        print("Saving check log to {}".format(log))

    def repair(self, save_path, img_dir=None):
        """
        Repair annotations
        :param save_path (str) : save the repaired coco address
        :param img_dir (str) : picture save folder, if None, do not check whether the picture exists
        """
        new_img_id = 0
        new_ano_id = 0
        new_coco = COCO_FORMAT
        cat_ids = self.coco.getCatIds()
        img_ids = self.coco.getImgIds()
        for ig_i in img_ids:
            img_info = self.coco.loadImgs(ig_i)[0]
            img_h = img_info["height"]
            img_w = img_info["width"]
            img_name = img_info["file_name"]
            if img_dir:
                img_path = os.path.join(img_dir, img_name)
                if not os.path.exists(img_path):
                    print("img path {} not exists!".format(img_path))
                    continue
            ano_ids = self.coco.getAnnIds(ig_i)
            if len(ano_ids) == 0:
                print("img id {} have't annotations!".format(ig_i))
                continue
            flag = 0
            for a_i in ano_ids:
                ano_info = self.coco.loadAnns(a_i)[0]
                cat_id = ano_info["category_id"]
                if cat_id not in cat_ids:
                    print("img id: {} ano id: {} category id: {} result: No annotation information!".format(ig_i, a_i,cat_id))
                    continue
                xywh = ano_info["bbox"]
                xywh = repairBox(xywh, img_h, img_w)
                ano_info["bbox"] = xywh
                ano_info["image_id"] = new_img_id
                ano_info["id"] = new_ano_id
                new_coco["annotations"].append(ano_info)
                new_ano_id += 1
                flag += 1
            if flag != 0:
                img_info["id"] = new_img_id
                new_coco["images"].append(img_info)
                new_img_id += 1
        new_coco["categories"] = self.coco.loadCats(cat_ids)
        with open(save_path, "w") as f:
            json.dump(new_coco, f, cls=MyEncoder)
            print("Complete data conversion to {}".format(save_path))


if __name__=="__main__":
    coco = "coco.json"
    repair_coco = "repair_coco.json"
    img_dir = "result"
    img_id = 2
    check = Checker(coco)

    check.drawImg(img_id)

    check.check()

    check.repair(repair_coco)











#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: transfer.py
# Author: Rayson.Shi
# Mail: raysonshi@qq.com
# Created Time:  2022-8-5 23:17:34
#############################################
from pycocotools.coco import COCO
import xml.etree.ElementTree as ET
from tqdm import tqdm
from cocoutils.utils import *


class Transfer:
    def __init__(self):
        super(Transfer, self).__init__()
        """
        Only detection format conversion is supported
        """
        self.coco_data = COCO_FORMAT
    
    def custom(self, dst_coco, img_abs_paths, cats, bboxes):
        """
        Custom dataset conversion to coco format
        :param dst_coco (str) : the path of Coco format JSON file to save
        :param img_abs_pathsh (list) : picture absolute address list, ["path/to/img1.jpg", "path/to/img2.jpg",...]
        :param img_abs_pathsh (list) : category name list, [["cat1","cat1","cat2"], ["cat2"],...]
        :param bboxes (list)  : xywh list, [[xywh1,xywh2,xyw3], [xywh4],...]
        The three must match one by one
        """
        assert len(cats) == len(bboxes), "The number of categories and Bboxes must be equal!"
        img_id, ano_id, cats_name = 0, 0, []
        for i, ig_p in enumerate(img_abs_paths):
            img_name = os.path.basename(ig_p)
            img = cv2.imread(ig_p)
            h, w, _ = img.shape
            img_sample = setImageSample(img_id, img_name, h, w)
            for j, (cat, xywh) in enumerate(zip(cats[i], bboxes[i])):
                cat_id, cat_sample = setCategorySample(cat, cats_name)
                if cat_sample:
                    self.coco_data["categories"].append(cat_sample)
                else:
                    ano_sample = setAnnotationSample(ano_id, img_id, cat_id, xywh)
                self.coco_data["annotations"].append(ano_sample)
                ano_id += 1
            self.coco_data["images"].append(img_sample)
            img_id += 1
        with open(dst_coco, "w") as f2:
            json.dump(self.coco_data, f2, cls=MyEncoder)
        print("Complete data conversion to {}".format(dst_coco))

    def lableme2coco(self, dst_coco, labelme_dir):
        """
        Labelme format to coco format
        :param dst_coco (str) : the path of Coco format JSON file to save
        :param labelme_dir (str) : JSON file folder
        """
        img_id, ano_id, cat_names = 0, 0, []
        js_paths = [os.path.join(labelme_dir, x) for x in os.listdir(labelme_dir) if x.endswith(".json")]
        for js_p in tqdm(js_paths):
            with open(js_p, "r") as f:
                data = json.load(f)
            img_name = data["imagePath"]
            img_h = data["imageHeight"]
            img_w = data["imageWidth"]
            img_sample = setImageSample(img_id, img_name, img_h, img_w)
            labelme_anos = data["shapes"]
            if len(labelme_anos) == 0:
                continue
            flag = ano_id
            for ano in labelme_anos:
                if ano["shape_type"] != "rectangle":
                    print("{} format not supported!".format(ano["shape_type"]))
                    continue
                cat = ano["label"]
                cat_id, cat_sample = setCategorySample(cat, cat_names)
                if cat_sample:
                    self.coco_data["categories"].append(cat_sample)
                points = ano["points"]
                xyxy = points2xyxy(points)
                xyxy = checkXyxy(xyxy, img_w, img_h)
                if xyxy is None:
                    continue
                xywh = xyxy2xywh(xyxy)
                anos_sample = setAnnotationSample(ano_id, img_id, cat_id, xywh)
                self.coco_data["annotations"].append(anos_sample)
                ano_id += 1
            if flag == ano_id:
                continue
            self.coco_data["images"].append(img_sample)
            img_id += 1
        with open(dst_coco, "w") as f2:
            json.dump(self.coco_data, f2, cls=MyEncoder)
        print("Complete data conversion to {}".format(dst_coco))

    def voc2coco(self, dst_coco, xml_dir):
        """
        voc format to coco format
        :param dst_coco (str) : the path of Coco format JSON file to save
        :param xml_dir (str) : XML file folder
        """
        img_id, ano_id, cat_names = 0, 0, []
        xml_names = [x for x in os.listdir(xml_dir) if x.endswith(".xml")]
        for name in tqdm(xml_names):
            xml_path = os.path.join(xml_dir, name)
            tree = ET.parse(xml_path)
            root = tree.getroot()
            img_name = getAndCheck(root, "filename", 1).text
            size = getAndCheck(root, 'size', 1)
            img_w = int(getAndCheck(size, 'width', 1).text)
            img_h = int(getAndCheck(size, 'height', 1).text)
            img_sample = setImageSample(img_id, img_name, img_h, img_w)
            flag = ano_id
            for obj in getAndCheck(root, 'object'):
                cat = getAndCheck(obj, 'name', 1).text
                cat_id, cat_sample = setCategorySample(cat, cat_names)
                if cat_sample:
                    self.coco_data["categories"].append(cat_sample)
                bndbox = getAndCheck(obj, 'bndbox', 1)
                x_min = int(float(getAndCheck(bndbox, 'xmin', 1).text))
                y_min = int(float(getAndCheck(bndbox, 'ymin', 1).text))
                x_max = int(float(getAndCheck(bndbox, 'xmax', 1).text))
                y_max = int(float(getAndCheck(bndbox, 'ymax', 1).text))
                xyxy = [x_min, y_min, x_max, y_max]
                xyxy = checkXyxy(xyxy, img_w, img_h)
                if xyxy is None:
                    continue
                xywh = xyxy2xywh(xyxy)
                ano_sample = setAnnotationSample(ano_id, img_id, cat_id, xywh)
                self.coco_data["annotations"].append(ano_sample)
                ano_id += 1
            if flag == ano_id:
                continue
            self.coco_data["images"].append(img_sample)
            img_id += 1
        with open(dst_coco, "w") as f:
            json.dump(self.coco_data, f, cls=MyEncoder)
        print("Complete data conversion to {}".format(dst_coco))

    def yolo2coco(self, dst_coco, txt_dir,  img_dir, cat_names=None):
        """
        yolo txt format to coco format
        :param dst_coco (str) : the path of Coco format JSON file to save
        :param txt_dir (str) : txt file folder
        :param img_dir (str) : picture folder
        :param cat_names (list) : specify category name, if None, use index as name
        """
        img_id, ano_id = 0, 0
        cat_names = [] if (cat_names is None) else cat_names
        cat_ids = []
        txt_names = [x for x in os.listdir(txt_dir) if x.endswith(".txt")]
        for txt_n in tqdm(txt_names):
            txt_p = os.path.join(txt_dir, txt_n)
            img_name = txt_n.replace(".txt", ".jpg")
            img_p = os.path.join(img_dir, img_name)
            if not os.path.exists(img_p):
                print("img path: {} not find!".format(img_p))
                img_p = img_p.replace(".jpg", ".png")
                print("try to find .png format!")
                if not os.path.exists(img_p):
                    print("img path: {} not find!".format(img_p))
                    continue
            img = cv2.imread(img_p)
            img_h, img_w, _ = img.shape
            img_sample = setImageSample(img_id, img_name, img_h,img_w)
            flag = ano_id
            with open(txt_p, "r") as f:
                anos = f.readlines()
                for ano in anos:
                    ano = ano.replace("\n", "").split(" ")
                    cat_id, yolo_box = int(float(ano[0])), list(map(float, ano[1:]))
                    if cat_id not in cat_ids:
                        cat_ids.append(cat_id)
                    xyxy = yoloBox2xyxy(yolo_box, img_h, img_w)
                    xyxy = checkXyxy(xyxy, img_w, img_h)
                    if xyxy is None:
                        continue
                    xywh = xyxy2xywh(xyxy)
                    ano_sample = setAnnotationSample(ano_id, img_id, cat_id, xywh)
                    self.coco_data["annotations"].append(ano_sample)
                    ano_id += 1
                if flag == ano_id:
                    continue
                self.coco_data["images"].append(img_sample)
                img_id += 1
        if len(cat_names) != 0:
            for i, c_n in enumerate(cat_names):
                _, cat_sample = setCategorySample(c_n, [])
                cat_sample["id"] = i
                self.coco_data["categories"].append(cat_sample)
        else:
            for c_i in cat_ids:
                _, cat_sample = setCategorySample(str(c_i), cat_names)
                self.coco_data["categories"].append(cat_sample)
        with open(dst_coco, "w") as f:
            json.dump(self.coco_data, f, cls=MyEncoder)
        print("Complete data conversion to {}".format(dst_coco))

    def mergeCoco(self, dst_coco, main_coco, second_coco):
        """
        Merge two coco datasets
        :param dst_coco (str) : the path of Coco format JSON file to save
        :param main_coco (str) : the path of Coco, as head
        :param second_coco (str) : the path of Coco, as tail
        """
        with open(main_coco, "r") as f1:
            main_data = json.load(f1)
        second_data = COCO(second_coco)
        main_imgs = main_data["images"]
        main_imgs_num = len(main_imgs)
        main_anos = main_data["annotations"]
        main_anos_num = len(main_anos)
        main_cat_names = [x["name"] for x in main_data["categories"]]
        cat_img_id = main_imgs[-1]["id"] + 1 if main_imgs_num < main_imgs[-1]["id"] else main_imgs_num
        cat_ano_id = main_anos[-1]["id"] + 1 if main_anos_num < main_imgs[-1]["id"] else main_anos_num
        second_cat_names = [x["name"] for x in second_data.loadCats(second_data.getCatIds())]
        t_img_id = 0
        t_ano_id = 0
        second_imgs_ids = second_data.getImgIds()
        for second_ig_id in second_imgs_ids:
            second_img_info = second_data.loadImgs(second_ig_id)[0]
            add_img_id = cat_img_id + t_img_id
            second_img_info["id"] = add_img_id
            second_ano_ids = second_data.getAnnIds(second_ig_id)
            if len(second_ano_ids) == 0:
                continue
            for second_ano_id in second_ano_ids:
                second_ano_info = second_data.loadAnns(second_ano_id)[0]
                second_cat_id = second_ano_info["category_id"]
                second_cat = second_cat_names[second_cat_id]
                add_cat_id, cat_sample = setCategorySample(second_cat, main_cat_names)
                if cat_sample:
                    main_data["categories"].append(cat_sample)
                add_ano_di = cat_ano_id + t_ano_id
                second_ano_info["id"] = add_ano_di
                second_ano_info["image_id"] = add_img_id
                second_ano_info["category_id"] = add_cat_id
                main_data["annotations"].append(second_ano_info)
                t_ano_id += 1
            main_data["images"].append(second_img_info)
            t_img_id += 1

        self.coco_data["images"] = main_data["images"]
        self.coco_data["annotations"] = main_data["annotations"]
        self.coco_data["categories"] = main_data["categories"]
        with open(dst_coco, "w") as f:
            json.dump(self.coco_data, f, cls=MyEncoder)
        print("Complete data conversion to {}".format(dst_coco))


if __name__=="__main__":
    labelme_dir = r"test"
    dst_coco = "coco.json"
    transfer = Transfer
    transfer.lableme2coco(dst_coco, labelme_dir)


#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: setup.py
# Author: Rayson.Shi
# Mail: raysonshi@qq.com
# Created Time:  2022-8-5 23:17:34
#############################################
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="cocoutils-cv",
    version="0.0.2",
    description=("Microsoft COCO data set transformation analysis utils"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="RaysonShi",
    author_email="raysonshi@qq.com",
    url="https://github.com/Rayson2020-8/cocodataset-utils",
    license="MIT Licence",
    packages=find_packages(),
    install_requires=[
        'numpy',
        'pycocotools',
        'matplotlib',
        'tqdm',
        'opencv-python',
    ],
    classifiers=[
                  "Programming Language :: Python :: 3",
                  "License :: OSI Approved :: MIT License",
                  "Operating System :: OS Independent",
              ]
)
